# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gtirb/proto/IR.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gtirb.proto import AuxData_pb2 as gtirb_dot_proto_dot_AuxData__pb2
from gtirb.proto import CFG_pb2 as gtirb_dot_proto_dot_CFG__pb2
from gtirb.proto import Module_pb2 as gtirb_dot_proto_dot_Module__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='gtirb/proto/IR.proto',
  package='gtirb.proto',
  syntax='proto3',
  serialized_pb=_b('\n\x14gtirb/proto/IR.proto\x12\x0bgtirb.proto\x1a\x19gtirb/proto/AuxData.proto\x1a\x15gtirb/proto/CFG.proto\x1a\x18gtirb/proto/Module.proto\"\x96\x02\n\x02IR\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12$\n\x07modules\x18\x03 \x03(\x0b\x32\x13.gtirb.proto.Module\x12.\n\x08\x61ux_data\x18\x05 \x03(\x0b\x32\x1c.gtirb.proto.IR.AuxDataEntry\x12\x0f\n\x07version\x18\x06 \x01(\r\x12\x1d\n\x03\x63\x66g\x18\x07 \x01(\x0b\x32\x10.gtirb.proto.CFG\x1a\x44\n\x0c\x41uxDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.gtirb.proto.AuxData:\x02\x38\x01J\x04\x08\x02\x10\x03J\x04\x08\x04\x10\x05R\x06tablesR\x0emain_module_idR\x12\x61ux_data_containerB\x1c\n\x1a\x63om.grammatech.gtirb.protob\x06proto3')
  ,
  dependencies=[gtirb_dot_proto_dot_AuxData__pb2.DESCRIPTOR,gtirb_dot_proto_dot_CFG__pb2.DESCRIPTOR,gtirb_dot_proto_dot_Module__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_IR_AUXDATAENTRY = _descriptor.Descriptor(
  name='AuxDataEntry',
  full_name='gtirb.proto.IR.AuxDataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='gtirb.proto.IR.AuxDataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='gtirb.proto.IR.AuxDataEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=268,
  serialized_end=336,
)

_IR = _descriptor.Descriptor(
  name='IR',
  full_name='gtirb.proto.IR',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='gtirb.proto.IR.uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='modules', full_name='gtirb.proto.IR.modules', index=1,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aux_data', full_name='gtirb.proto.IR.aux_data', index=2,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='version', full_name='gtirb.proto.IR.version', index=3,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='cfg', full_name='gtirb.proto.IR.cfg', index=4,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_IR_AUXDATAENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=114,
  serialized_end=392,
)

_IR_AUXDATAENTRY.fields_by_name['value'].message_type = gtirb_dot_proto_dot_AuxData__pb2._AUXDATA
_IR_AUXDATAENTRY.containing_type = _IR
_IR.fields_by_name['modules'].message_type = gtirb_dot_proto_dot_Module__pb2._MODULE
_IR.fields_by_name['aux_data'].message_type = _IR_AUXDATAENTRY
_IR.fields_by_name['cfg'].message_type = gtirb_dot_proto_dot_CFG__pb2._CFG
DESCRIPTOR.message_types_by_name['IR'] = _IR

IR = _reflection.GeneratedProtocolMessageType('IR', (_message.Message,), dict(

  AuxDataEntry = _reflection.GeneratedProtocolMessageType('AuxDataEntry', (_message.Message,), dict(
    DESCRIPTOR = _IR_AUXDATAENTRY,
    __module__ = 'gtirb.proto.IR_pb2'
    # @@protoc_insertion_point(class_scope:gtirb.proto.IR.AuxDataEntry)
    ))
  ,
  DESCRIPTOR = _IR,
  __module__ = 'gtirb.proto.IR_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.IR)
  ))
_sym_db.RegisterMessage(IR)
_sym_db.RegisterMessage(IR.AuxDataEntry)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\032com.grammatech.gtirb.proto'))
_IR_AUXDATAENTRY.has_options = True
_IR_AUXDATAENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gtirb/proto/CodeBlock.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='gtirb/proto/CodeBlock.proto',
  package='gtirb.proto',
  syntax='proto3',
  serialized_pb=_b('\n\x1bgtirb/proto/CodeBlock.proto\x12\x0bgtirb.proto\"K\n\tCodeBlock\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x0c\n\x04size\x18\x03 \x01(\x04\x12\x13\n\x0b\x64\x65\x63ode_mode\x18\x04 \x01(\x04J\x04\x08\x02\x10\x03R\x07\x61\x64\x64ressB\x1c\n\x1a\x63om.grammatech.gtirb.protob\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_CODEBLOCK = _descriptor.Descriptor(
  name='CodeBlock',
  full_name='gtirb.proto.CodeBlock',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='gtirb.proto.CodeBlock.uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='size', full_name='gtirb.proto.CodeBlock.size', index=1,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='decode_mode', full_name='gtirb.proto.CodeBlock.decode_mode', index=2,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=119,
)

DESCRIPTOR.message_types_by_name['CodeBlock'] = _CODEBLOCK

CodeBlock = _reflection.GeneratedProtocolMessageType('CodeBlock', (_message.Message,), dict(
  DESCRIPTOR = _CODEBLOCK,
  __module__ = 'gtirb.proto.CodeBlock_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.CodeBlock)
  ))
_sym_db.RegisterMessage(CodeBlock)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\032com.grammatech.gtirb.proto'))
# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gtirb/proto/SymbolicExpression.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='gtirb/proto/SymbolicExpression.proto',
  package='gtirb.proto',
  syntax='proto3',
  serialized_pb=_b('\n$gtirb/proto/SymbolicExpression.proto\x12\x0bgtirb.proto\"4\n\rSymStackConst\x12\x0e\n\x06offset\x18\x01 \x01(\x05\x12\x13\n\x0bsymbol_uuid\x18\x02 \x01(\x0c\"3\n\x0cSymAddrConst\x12\x0e\n\x06offset\x18\x01 \x01(\x03\x12\x13\n\x0bsymbol_uuid\x18\x02 \x01(\x0c\"X\n\x0bSymAddrAddr\x12\r\n\x05scale\x18\x01 \x01(\x03\x12\x0e\n\x06offset\x18\x02 \x01(\x03\x12\x14\n\x0csymbol1_uuid\x18\x03 \x01(\x0c\x12\x14\n\x0csymbol2_uuid\x18\x04 \x01(\x0c\"\xb0\x01\n\x12SymbolicExpression\x12\x31\n\x0bstack_const\x18\x01 \x01(\x0b\x32\x1a.gtirb.proto.SymStackConstH\x00\x12/\n\naddr_const\x18\x02 \x01(\x0b\x32\x19.gtirb.proto.SymAddrConstH\x00\x12-\n\taddr_addr\x18\x03 \x01(\x0b\x32\x18.gtirb.proto.SymAddrAddrH\x00\x42\x07\n\x05valueB\x1c\n\x1a\x63om.grammatech.gtirb.protob\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_SYMSTACKCONST = _descriptor.Descriptor(
  name='SymStackConst',
  full_name='gtirb.proto.SymStackConst',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='offset', full_name='gtirb.proto.SymStackConst.offset', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symbol_uuid', full_name='gtirb.proto.SymStackConst.symbol_uuid', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=105,
)


_SYMADDRCONST = _descriptor.Descriptor(
  name='SymAddrConst',
  full_name='gtirb.proto.SymAddrConst',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='offset', full_name='gtirb.proto.SymAddrConst.offset', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symbol_uuid', full_name='gtirb.proto.SymAddrConst.symbol_uuid', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=107,
  serialized_end=158,
)


_SYMADDRADDR = _descriptor.Descriptor(
  name='SymAddrAddr',
  full_name='gtirb.proto.SymAddrAddr',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='scale', full_name='gtirb.proto.SymAddrAddr.scale', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='offset', full_name='gtirb.proto.SymAddrAddr.offset', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symbol1_uuid', full_name='gtirb.proto.SymAddrAddr.symbol1_uuid', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symbol2_uuid', full_name='gtirb.proto.SymAddrAddr.symbol2_uuid', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=160,
  serialized_end=248,
)


_SYMBOLICEXPRESSION = _descriptor.Descriptor(
  name='SymbolicExpression',
  full_name='gtirb.proto.SymbolicExpression',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='stack_const', full_name='gtirb.proto.SymbolicExpression.stack_const', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='addr_const', full_name='gtirb.proto.SymbolicExpression.addr_const', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='addr_addr', full_name='gtirb.proto.SymbolicExpression.addr_addr', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='gtirb.proto.SymbolicExpression.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=251,
  serialized_end=427,
)

_SYMBOLICEXPRESSION.fields_by_name['stack_const'].message_type = _SYMSTACKCONST
_SYMBOLICEXPRESSION.fields_by_name['addr_const'].message_type = _SYMADDRCONST
_SYMBOLICEXPRESSION.fields_by_name['addr_addr'].message_type = _SYMADDRADDR
_SYMBOLICEXPRESSION.oneofs_by_name['value'].fields.append(
  _SYMBOLICEXPRESSION.fields_by_name['stack_const'])
_SYMBOLICEXPRESSION.fields_by_name['stack_const'].containing_oneof = _SYMBOLICEXPRESSION.oneofs_by_name['value']
_SYMBOLICEXPRESSION.oneofs_by_name['value'].fields.append(
  _SYMBOLICEXPRESSION.fields_by_name['addr_const'])
_SYMBOLICEXPRESSION.fields_by_name['addr_const'].containing_oneof = _SYMBOLICEXPRESSION.oneofs_by_name['value']
_SYMBOLICEXPRESSION.oneofs_by_name['value'].fields.append(
  _SYMBOLICEXPRESSION.fields_by_name['addr_addr'])
_SYMBOLICEXPRESSION.fields_by_name['addr_addr'].containing_oneof = _SYMBOLICEXPRESSION.oneofs_by_name['value']
DESCRIPTOR.message_types_by_name['SymStackConst'] = _SYMSTACKCONST
DESCRIPTOR.message_types_by_name['SymAddrConst'] = _SYMADDRCONST
DESCRIPTOR.message_types_by_name['SymAddrAddr'] = _SYMADDRADDR
DESCRIPTOR.message_types_by_name['SymbolicExpression'] = _SYMBOLICEXPRESSION

SymStackConst = _reflection.GeneratedProtocolMessageType('SymStackConst', (_message.Message,), dict(
  DESCRIPTOR = _SYMSTACKCONST,
  __module__ = 'gtirb.proto.SymbolicExpression_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.SymStackConst)
  ))
_sym_db.RegisterMessage(SymStackConst)

SymAddrConst = _reflection.GeneratedProtocolMessageType('SymAddrConst', (_message.Message,), dict(
  DESCRIPTOR = _SYMADDRCONST,
  __module__ = 'gtirb.proto.SymbolicExpression_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.SymAddrConst)
  ))
_sym_db.RegisterMessage(SymAddrConst)

SymAddrAddr = _reflection.GeneratedProtocolMessageType('SymAddrAddr', (_message.Message,), dict(
  DESCRIPTOR = _SYMADDRADDR,
  __module__ = 'gtirb.proto.SymbolicExpression_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.SymAddrAddr)
  ))
_sym_db.RegisterMessage(SymAddrAddr)

SymbolicExpression = _reflection.GeneratedProtocolMessageType('SymbolicExpression', (_message.Message,), dict(
  DESCRIPTOR = _SYMBOLICEXPRESSION,
  __module__ = 'gtirb.proto.SymbolicExpression_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.SymbolicExpression)
  ))
_sym_db.RegisterMessage(SymbolicExpression)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\032com.grammatech.gtirb.proto'))
# @@protoc_insertion_point(module_scope)



"""General utilities usable by any other GTIRB submoudle."""

import itertools
import typing

K = typing.TypeVar("K")
V = typing.TypeVar("V")
T = typing.TypeVar("T")


class DictLike(typing.Generic[K, V]):
    """Any value that can be passed to the constructor of ``dict``;
    that is, a mapping or iterable yielding key-value tuples.
    """

    # NOTE: this should inherit from:
    # typing.Union[
    #     typing.Mapping[K, V],
    #     typing.Iterable[typing.Tuple[K, V]],
    # ]
    # but cannot, due to metaclass conflicts.


class ListWrapper(typing.MutableSequence[T]):
    def __init__(self, *args):
        self._data = []
        for v in list(*args):
            self.append(v)

    # begin functions for ABC
    def __getitem__(self, i):
        return self._data[i]

    def __setitem__(self, i, v):
        self._data[i] = v

    def __delitem__(self, i):
        del self._data[i]

    def __len__(self):
        return len(self._data)

    def insert(self, i, v):
        return self._data.insert(i, v)

    # The version of typing.py which comes with python 3.5.2 doesn't provide
    # definitions for append or remove on MutableList, so we have to do it
    # ourselves.
    def append(self, v):
        self.insert(len(self), v)

    def remove(self, v):
        del self[self._data.index(v)]

    # extend is not in every version of Python 3, so list wrapper adds it here
    # itself.
    def extend(self, other):
        for v in other:
            self.append(v)

    # end functions for ABC
    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)


class SetWrapper(typing.MutableSet[T]):
    def __init__(self, *args):
        self._data = set()
        for v in set(*args):
            self.add(v)

    # begin functions for ABC
    def __contains__(self, v):
        return v in self._data

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def add(self, v):
        return self._data.add(v)

    def discard(self, v):
        return self._data.discard(v)

    # end functions for ABC

    # The version of typing.py which comes with python 3.5.2 doesn't provide
    # definitions for __or__ or clear on MutableSet, so we have to do it
    # ourselves.
    def __or__(self, other):
        return self._data | other

    def clear(self):
        self._data.clear()

    # For whatever reason, update isn't included as part of abc.MutableSet.
    def update(self, *others):
        for other in others:
            for v in other:
                self.add(v)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)


class DictWrapper(typing.MutableMapping[K, V]):
    def __init__(self, *args):
        self._data = {}
        for i, v in dict(*args).items():
            self[i] = v

    # begin functions for ABC
    def __getitem__(self, i):
        return self._data[i]

    def __setitem__(self, i, v):
        self._data[i] = v

    def __delitem__(self, i):
        del self._data[i]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    # end functions for ABC
    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)


def get_desired_range(addrs):
    # type: (typing.Union[int, range]) -> range
    if isinstance(addrs, int):
        return range(addrs, addrs + 1)
    else:
        return addrs


def nodes_on(
    nodes,  # type: typing.Iterable[T]
    addrs,  # type: typing.Union[int, range]
):
    # type: (...) -> typing.Iterable[T]
    desired_range = get_desired_range(addrs)
    for node in nodes:
        node_addr = node.address
        if node_addr is not None:
            node_range = range(node_addr, node_addr + node.size)
            if range(
                max(desired_range.start, node_range.start),
                min(desired_range.stop, node_range.stop),
            ):
                yield node


def nodes_at(
    nodes,  # type: typing.Iterable[T]
    addrs,  # type: typing.Union[int, range]
):
    # type: (...) -> typing.Iterable[T]
    desired_range = get_desired_range(addrs)
    for node in nodes:
        node_addr = node.address
        if node_addr is not None and node_addr in desired_range:
            yield node


def symbolic_expressions_at(
    nodes,  # type: typing.Iterable
    addrs,  # type: typing.Union[int, range]
):
    # type: (...) -> typing.Iterable[typing.Tuple]
    return itertools.chain.from_iterable(
        node.symbolic_expressions_at(addrs) for node in nodes
    )



import typing
from uuid import UUID

from .node import Node
from .proto import SymbolicExpression_pb2
from .symbol import Symbol


class SymAddrAddr:
    """Represents a symbolic expression of the form
    "(Sym1 - Sym2) / Scale + Offset".

    :ivar ~.scale: Constant scale factor.
    :ivar ~.offset: Constant offset.
    :ivar ~.symbol1: Symbol representing the base address.
    :ivar ~.symbol2: Symbol to subtract from ``symbol1``.
    """

    def __init__(
        self,
        scale,  # type: int
        offset,  # type: int
        symbol1,  # type: Symbol
        symbol2,  # type: Symbol
    ):
        # type: (...) -> None
        """
        :param scale: Constant scale factor.
        :param offset: Constant offset.
        :param symbol1: Symbol representing the base address.
        :param symbol2: Symbol to subtract from ``symbol1``.
        """

        self.scale = scale  # type: int
        self.offset = offset  # type: int
        self.symbol1 = symbol1  # type: Symbol
        self.symbol2 = symbol2  # type: Symbol

    @classmethod
    def _from_protobuf(
        cls,
        proto_symaddraddr,  # type: SymbolicExpression_pb2.SymAddrAddr
        get_by_uuid,  # type: typing.Callable[[UUID], Node]
    ):
        # type: (...) -> SymAddrAddr
        symbol1 = get_by_uuid(UUID(bytes=proto_symaddraddr.symbol1_uuid))
        symbol2 = get_by_uuid(UUID(bytes=proto_symaddraddr.symbol2_uuid))
        return cls(
            proto_symaddraddr.scale, proto_symaddraddr.offset, symbol1, symbol2
        )

    def _to_protobuf(self):
        # type: () -> SymbolicExpression_pb2.SymAddrAddr
        proto_symaddraddr = SymbolicExpression_pb2.SymAddrAddr()
        proto_symaddraddr.scale = self.scale
        proto_symaddraddr.offset = self.offset
        proto_symaddraddr.symbol1_uuid = self.symbol1.uuid.bytes
        proto_symaddraddr.symbol2_uuid = self.symbol2.uuid.bytes
        return proto_symaddraddr

    def __eq__(self, other):
        # type: (typing.Any) -> bool
        if not isinstance(other, SymAddrAddr):
            return False
        return (
            self.scale == other.scale
            and self.offset == other.offset
            and self.symbol1.uuid == other.symbol1.uuid
            and self.symbol2.uuid == other.symbol2.uuid
        )

    def __hash__(self):
        # type: () -> int
        return hash(
            (self.offset, self.scale, self.symbol1.uuid, self.symbol2.uuid)
        )

    def __repr__(self):
        # type: () -> str
        return (
            "SymAddrAddr("
            "scale={scale!r}, "
            "offset={offset!r}, "
            "symbol1={symbol1!r}, "
            "symbol2={symbol2!r}, "
            ")".format(**self.__dict__)
        )

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, SymAddrAddr):
            return False
        return (
            self.scale == other.scale
            and self.offset == other.offset
            and self.symbol1.deep_eq(other.symbol1)
            and self.symbol2.deep_eq(other.symbol2)
        )


class SymAddrConst:
    """Represents a symbolic expression of the form "Sym + Offset".

    :ivar ~.offset: Constant offset.
    :ivar ~.symbol: Symbol representing an address.
    """

    def __init__(self, offset, symbol):
        # type: (int,Symbol) -> None
        """
        :param offset: Constant offset.
        :param symbol: Symbol representing an address.
        """

        self.offset = offset  # type: int
        self.symbol = symbol  # type: Symbol

    @classmethod
    def _from_protobuf(
        cls,
        proto_symaddrconst,  # type: SymbolicExpression_pb2.SymAddrConst
        get_by_uuid,  # type: typing.Callable[[UUID], Node]
    ):
        # type: (...) -> SymAddrConst
        symbol = get_by_uuid(UUID(bytes=proto_symaddrconst.symbol_uuid))
        return cls(proto_symaddrconst.offset, symbol)

    def _to_protobuf(self):
        # type: () -> SymbolicExpression_pb2.SymAddrConst
        proto_symaddrconst = SymbolicExpression_pb2.SymAddrConst()
        proto_symaddrconst.offset = self.offset
        if self.symbol is not None:
            proto_symaddrconst.symbol_uuid = self.symbol.uuid.bytes
        return proto_symaddrconst

    def __eq__(self, other):
        # type: (typing.Any) -> bool
        if not isinstance(other, SymAddrConst):
            return False
        return (
            self.offset == other.offset
            and self.symbol.uuid == other.symbol.uuid
        )

    def __hash__(self):
        # type: () -> int
        return hash((self.offset, self.symbol.uuid))

    def __repr__(self):
        # type: () -> str
        return (
            "SymAddrConst("
            "offset={offset!r}, "
            "symbol={symbol!r}, "
            ")".format(**self.__dict__)
        )

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, SymAddrConst):
            return False
        return self.offset == other.offset and self.symbol.deep_eq(
            other.symbol
        )


class SymStackConst:
    """Represents a symbolic expression of the form "Sym + Offset",
    representing an offset from a stack variable.

    :ivar ~.offset: Constant offset.
    :ivar ~.symbol: Symbol representing a stack variable.
    """

    def __init__(self, offset, symbol):
        # type: (int,Symbol) -> None
        """
        :param offset: Constant offset.
        :param symbol: Symbol representing a stack variable.
        """

        self.offset = offset  # type: int
        self.symbol = symbol  # type: Symbol

    @classmethod
    def _from_protobuf(
        cls,
        proto_symstackconst,  # type: SymbolicExpression_pb2.SymStackConst
        get_by_uuid,  # type: typing.Callable[[UUID], Node]
    ):
        # type: (...) -> SymStackConst
        symbol = get_by_uuid(UUID(bytes=proto_symstackconst.symbol_uuid))
        return cls(proto_symstackconst.offset, symbol)

    def _to_protobuf(self):
        # type: () -> SymbolicExpression_pb2.SymStackConst
        proto_symstackconst = SymbolicExpression_pb2.SymStackConst()
        proto_symstackconst.offset = self.offset
        if self.symbol is not None:
            proto_symstackconst.symbol_uuid = self.symbol.uuid.bytes
        return proto_symstackconst

    def __eq__(self, other):
        # type: (typing.Any) -> bool
        if not isinstance(other, SymStackConst):
            return False
        return (
            self.offset == other.offset
            and self.symbol.uuid == other.symbol.uuid
        )

    def __hash__(self):
        # type: () -> int
        return hash((self.offset, self.symbol.uuid))

    def __repr__(self):
        # type: () -> str
        return (
            "SymStackConst("
            "offset={offset!r}, "
            "symbol={symbol!r}, "
            ")".format(**self.__dict__)
        )

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, SymStackConst):
            return False
        return self.offset == other.offset and self.symbol.deep_eq(
            other.symbol
        )


SymbolicExpression = typing.Union[SymAddrAddr, SymAddrConst, SymStackConst]
"""A type hint for any symbolic expression type."""



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gtirb/proto/Module.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gtirb.proto import AuxData_pb2 as gtirb_dot_proto_dot_AuxData__pb2
from gtirb.proto import ProxyBlock_pb2 as gtirb_dot_proto_dot_ProxyBlock__pb2
from gtirb.proto import Section_pb2 as gtirb_dot_proto_dot_Section__pb2
from gtirb.proto import Symbol_pb2 as gtirb_dot_proto_dot_Symbol__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='gtirb/proto/Module.proto',
  package='gtirb.proto',
  syntax='proto3',
  serialized_pb=_b('\n\x18gtirb/proto/Module.proto\x12\x0bgtirb.proto\x1a\x19gtirb/proto/AuxData.proto\x1a\x1cgtirb/proto/ProxyBlock.proto\x1a\x19gtirb/proto/Section.proto\x1a\x18gtirb/proto/Symbol.proto\"\xa9\x04\n\x06Module\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x13\n\x0b\x62inary_path\x18\x02 \x01(\t\x12\x16\n\x0epreferred_addr\x18\x03 \x01(\x04\x12\x14\n\x0crebase_delta\x18\x04 \x01(\x03\x12,\n\x0b\x66ile_format\x18\x05 \x01(\x0e\x32\x17.gtirb.proto.FileFormat\x12\x1d\n\x03isa\x18\x06 \x01(\x0e\x32\x10.gtirb.proto.ISA\x12\x0c\n\x04name\x18\x07 \x01(\t\x12$\n\x07symbols\x18\t \x03(\x0b\x32\x13.gtirb.proto.Symbol\x12(\n\x07proxies\x18\x10 \x03(\x0b\x32\x17.gtirb.proto.ProxyBlock\x12&\n\x08sections\x18\x0c \x03(\x0b\x32\x14.gtirb.proto.Section\x12\x32\n\x08\x61ux_data\x18\x11 \x03(\x0b\x32 .gtirb.proto.Module.AuxDataEntry\x12\x13\n\x0b\x65ntry_point\x18\x12 \x01(\x0c\x1a\x44\n\x0c\x41uxDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.gtirb.proto.AuxData:\x02\x38\x01J\x04\x08\x08\x10\tJ\x04\x08\n\x10\x0bJ\x04\x08\x0b\x10\x0cJ\x04\x08\r\x10\x0eJ\x04\x08\x0e\x10\x0fJ\x04\x08\x0f\x10\x10R\x0eimage_byte_mapR\x04\x64\x61taR\x11symbolic_operandsR\x12\x61ux_data_containerR\x06\x62locksR\x03\x63\x66g*|\n\nFileFormat\x12\x14\n\x10\x46ormat_Undefined\x10\x00\x12\x08\n\x04\x43OFF\x10\x01\x12\x07\n\x03\x45LF\x10\x02\x12\x06\n\x02PE\x10\x03\x12\x0e\n\nIdaProDb32\x10\x04\x12\x0e\n\nIdaProDb64\x10\x05\x12\t\n\x05XCOFF\x10\x06\x12\t\n\x05MACHO\x10\x07\x12\x07\n\x03RAW\x10\x08*\x86\x01\n\x03ISA\x12\x11\n\rISA_Undefined\x10\x00\x12\x08\n\x04IA32\x10\x01\x12\t\n\x05PPC32\x10\x02\x12\x07\n\x03X64\x10\x03\x12\x07\n\x03\x41RM\x10\x04\x12\x17\n\x13ValidButUnsupported\x10\x05\x12\t\n\x05PPC64\x10\x06\x12\t\n\x05\x41RM64\x10\x07\x12\n\n\x06MIPS32\x10\x08\x12\n\n\x06MIPS64\x10\tB\x1c\n\x1a\x63om.grammatech.gtirb.protob\x06proto3')
  ,
  dependencies=[gtirb_dot_proto_dot_AuxData__pb2.DESCRIPTOR,gtirb_dot_proto_dot_ProxyBlock__pb2.DESCRIPTOR,gtirb_dot_proto_dot_Section__pb2.DESCRIPTOR,gtirb_dot_proto_dot_Symbol__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_FILEFORMAT = _descriptor.EnumDescriptor(
  name='FileFormat',
  full_name='gtirb.proto.FileFormat',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='Format_Undefined', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COFF', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ELF', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PE', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IdaProDb32', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IdaProDb64', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='XCOFF', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MACHO', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAW', index=8, number=8,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=707,
  serialized_end=831,
)
_sym_db.RegisterEnumDescriptor(_FILEFORMAT)

FileFormat = enum_type_wrapper.EnumTypeWrapper(_FILEFORMAT)
_ISA = _descriptor.EnumDescriptor(
  name='ISA',
  full_name='gtirb.proto.ISA',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ISA_Undefined', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IA32', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPC32', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='X64', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARM', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ValidButUnsupported', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPC64', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARM64', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MIPS32', index=8, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MIPS64', index=9, number=9,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=834,
  serialized_end=968,
)
_sym_db.RegisterEnumDescriptor(_ISA)

ISA = enum_type_wrapper.EnumTypeWrapper(_ISA)
Format_Undefined = 0
COFF = 1
ELF = 2
PE = 3
IdaProDb32 = 4
IdaProDb64 = 5
XCOFF = 6
MACHO = 7
RAW = 8
ISA_Undefined = 0
IA32 = 1
PPC32 = 2
X64 = 3
ARM = 4
ValidButUnsupported = 5
PPC64 = 6
ARM64 = 7
MIPS32 = 8
MIPS64 = 9



_MODULE_AUXDATAENTRY = _descriptor.Descriptor(
  name='AuxDataEntry',
  full_name='gtirb.proto.Module.AuxDataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='gtirb.proto.Module.AuxDataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='gtirb.proto.Module.AuxDataEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=527,
  serialized_end=595,
)

_MODULE = _descriptor.Descriptor(
  name='Module',
  full_name='gtirb.proto.Module',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='gtirb.proto.Module.uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='binary_path', full_name='gtirb.proto.Module.binary_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='preferred_addr', full_name='gtirb.proto.Module.preferred_addr', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rebase_delta', full_name='gtirb.proto.Module.rebase_delta', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='file_format', full_name='gtirb.proto.Module.file_format', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='isa', full_name='gtirb.proto.Module.isa', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='gtirb.proto.Module.name', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symbols', full_name='gtirb.proto.Module.symbols', index=7,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='proxies', full_name='gtirb.proto.Module.proxies', index=8,
      number=16, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sections', full_name='gtirb.proto.Module.sections', index=9,
      number=12, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aux_data', full_name='gtirb.proto.Module.aux_data', index=10,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='entry_point', full_name='gtirb.proto.Module.entry_point', index=11,
      number=18, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_MODULE_AUXDATAENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=705,
)

_MODULE_AUXDATAENTRY.fields_by_name['value'].message_type = gtirb_dot_proto_dot_AuxData__pb2._AUXDATA
_MODULE_AUXDATAENTRY.containing_type = _MODULE
_MODULE.fields_by_name['file_format'].enum_type = _FILEFORMAT
_MODULE.fields_by_name['isa'].enum_type = _ISA
_MODULE.fields_by_name['symbols'].message_type = gtirb_dot_proto_dot_Symbol__pb2._SYMBOL
_MODULE.fields_by_name['proxies'].message_type = gtirb_dot_proto_dot_ProxyBlock__pb2._PROXYBLOCK
_MODULE.fields_by_name['sections'].message_type = gtirb_dot_proto_dot_Section__pb2._SECTION
_MODULE.fields_by_name['aux_data'].message_type = _MODULE_AUXDATAENTRY
DESCRIPTOR.message_types_by_name['Module'] = _MODULE
DESCRIPTOR.enum_types_by_name['FileFormat'] = _FILEFORMAT
DESCRIPTOR.enum_types_by_name['ISA'] = _ISA

Module = _reflection.GeneratedProtocolMessageType('Module', (_message.Message,), dict(

  AuxDataEntry = _reflection.GeneratedProtocolMessageType('AuxDataEntry', (_message.Message,), dict(
    DESCRIPTOR = _MODULE_AUXDATAENTRY,
    __module__ = 'gtirb.proto.Module_pb2'
    # @@protoc_insertion_point(class_scope:gtirb.proto.Module.AuxDataEntry)
    ))
  ,
  DESCRIPTOR = _MODULE,
  __module__ = 'gtirb.proto.Module_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.Module)
  ))
_sym_db.RegisterMessage(Module)
_sym_db.RegisterMessage(Module.AuxDataEntry)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\032com.grammatech.gtirb.proto'))
_MODULE_AUXDATAENTRY.has_options = True
_MODULE_AUXDATAENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
# @@protoc_insertion_point(module_scope)



import typing
from uuid import UUID

from .node import Node
from .proto import CodeBlock_pb2, DataBlock_pb2, ProxyBlock_pb2


class Block(Node):
    """The base class for blocks. Symbols may have references to any subclass
    of Block.
    """

    @property
    def references(self):
        # type: () -> typing.Iterable["Symbol"]
        """Get all the symbols that refer to this block."""

        raise NotImplementedError

    def _add_to_uuid_cache(self, cache):
        # type: (typing.Dict[UUID, Node]) -> None
        """Update the UUID cache when this node is added."""

        cache[self.uuid] = self

    def _remove_from_uuid_cache(self, cache):
        # type: (typing.Dict[UUID, Node]) -> None
        """Update the UUID cache when this node is removed."""

        del cache[self.uuid]


class ByteBlock(Block):
    """The base class for blocks that belong to a :class:`ByteInterval` and
    store their bytes there.

    :ivar ~.size: The size of the block in bytes.
    :ivar ~.offset: The offset from the beginning of the byte interval to which
        this block belongs. Multiple blocks in the same interval may have the
        same offset.
    """

    def __init__(
        self,
        *,
        size=0,  # type: int
        offset=0,  # type: int
        uuid=None  # type: typing.Optional[UUID]
    ):
        """
        :param size: The size of the data object in bytes.
        :param offset: The offset from the beginning of the byte interval to
            which this block belongs.
        :param uuid: The UUID of this ``ByteBlock``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        """

        super().__init__(uuid=uuid)
        self.size = size  # type: int
        self.offset = offset  # type: int
        self._byte_interval = None  # type: typing.Optional["ByteInterval"]

    @property
    def byte_interval(self):
        # type: () -> typing.Optional["ByteInterval"]
        """The :class:`ByteInterval` this block belongs to."""

        return self._byte_interval

    @byte_interval.setter
    def byte_interval(self, value):
        # type: (typing.Optional["ByteInterval"]) -> None
        if self._byte_interval is not None:
            self._byte_interval.blocks.discard(self)
        if value is not None:
            value.blocks.add(self)

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, ByteBlock):
            return False
        return (
            self.offset == other.offset
            and self.uuid == other.uuid
            and self.size == other.size
        )

    @property
    def contents(self):
        # type: () -> bytes
        """Get the bytes in this block."""

        if self.byte_interval is None:
            return b""
        return self.byte_interval.contents[
            self.offset : self.offset + self.size
        ]

    @property
    def address(self):
        # type: () -> typing.Optional[int]
        """Get the address of this block, or None if not present."""

        if self.byte_interval is None or self.byte_interval.address is None:
            return None
        return self.byte_interval.address + self.offset

    @property
    def references(self):
        if (
            self.byte_interval is None
            or self.byte_interval.section is None
            or self.byte_interval.section.module is None
        ):
            return ()
        return (
            s
            for s in self.byte_interval.section.module.symbols
            if s.referent == self
        )

    @property
    def section(self):
        # type: () -> "Section"
        """Get the section this node ultimately belongs to."""
        if self.byte_interval is None:
            return None
        return self.byte_interval.section

    @property
    def module(self):
        # type: () -> "Module"
        """Get the module this node ultimately belongs to."""
        if self.section is None:
            return None
        return self.section.module

    @property
    def ir(self):
        # type: () -> "IR"
        """Get the IR this node ultimately belongs to."""
        if self.module is None:
            return None
        return self.module.ir


class CfgNode(Block):
    """The base class for blocks that may appear as vertices in the CFG."""

    @property
    def incoming_edges(self):
        # type: () -> typing.Iterable["Edge"]
        """Get the edges that point to this CFG node."""

        raise NotImplementedError

    @property
    def outgoing_edges(self):
        # type: () -> typing.Iterable["Edge"]
        """Get the edges that start at this CFG node."""

        raise NotImplementedError


class DataBlock(ByteBlock):
    """Represents a data object, possibly symbolic."""

    def __init__(self, *, size=0, offset=0, uuid=None):
        # type: (int, int, typing.Optional[UUID]) -> None
        """
        :param size: The size of the data object in bytes.
        :param offset: The offset from the beginning of the byte interval to
            which this block belongs.
        :param uuid: The UUID of this ``DataBlock``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        """

        super().__init__(size=size, offset=offset, uuid=uuid)

    @classmethod
    def _decode_protobuf(
        cls,
        proto_dataobject,  # type: DataBlock_pb2.DataBlock
        uuid,  # type: uuid.UUID
        ir,  # type: typing.Optional["IR"]
    ):
        # type: (...) -> DataBlock
        b = cls(size=proto_dataobject.size, uuid=uuid)
        b._add_to_uuid_cache(ir._local_uuid_cache)
        return b

    def _to_protobuf(self):
        # type: () -> DataBlock_pb2.DataBlock
        proto_dataobject = DataBlock_pb2.DataBlock()
        proto_dataobject.uuid = self.uuid.bytes
        proto_dataobject.size = self.size
        return proto_dataobject

    def __repr__(self):
        # type: () -> str
        return (
            "DataBlock("
            "uuid={uuid!r}, "
            "size={size}, "
            "offset={offset}, "
            ")".format(**self.__dict__)
        )


class CodeBlock(ByteBlock, CfgNode):
    """A basic block in the binary.

    Does not directly store data bytes, which are kept in a
    :class:`ByteInterval`.

    :ivar ~.decode_mode: The decode mode of the block,
        used in some ISAs to differentiate between sub-ISAs
        (e.g. differentiating blocks written in ARM and Thumb).
    """

    def __init__(
        self,
        *,
        decode_mode=0,  # type: int
        size=0,  # type: int
        offset=0,  # type: int
        uuid=None  # type: typing.Optional[UUID]
    ):
        # type: (...) -> None
        """
        :param size: The length of the block in bytes.
        :param decode_mode: The decode mode of the block,
            used in some ISAs to differentiate between sub-ISAs
            (e.g. differentiating blocks written in ARM and Thumb).
            Defaults to 0.
        :param offset: The offset from the beginning of the byte interval to
            which this block belongs.
        :param uuid: The UUID of this ``CodeBlock``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        """

        super().__init__(size=size, offset=offset, uuid=uuid)
        self.decode_mode = decode_mode  # type: int

    @classmethod
    def _decode_protobuf(
        cls,
        proto_block,  # type: CodeBlock_pb2.CodeBlock
        uuid,  # type: UUID
        ir,  # type: typing.Optional["IR"]
    ):
        # type: (...) -> CodeBlock
        b = cls(
            decode_mode=proto_block.decode_mode,
            size=proto_block.size,
            uuid=uuid,
        )
        b._add_to_uuid_cache(ir._local_uuid_cache)
        return b

    def _to_protobuf(self):
        # type: () -> CodeBlock_pb2.CodeBlock
        proto_block = CodeBlock_pb2.CodeBlock()
        proto_block.uuid = self.uuid.bytes
        proto_block.size = self.size
        proto_block.decode_mode = self.decode_mode
        return proto_block

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, CodeBlock):
            return False
        return super().deep_eq(other) and self.decode_mode == other.decode_mode

    def __repr__(self):
        # type: () -> str
        return (
            "CodeBlock("
            "uuid={uuid!r}, "
            "size={size}, "
            "offset={offset}, "
            "decode_mode={decode_mode}, "
            ")".format(**self.__dict__)
        )

    @property
    def incoming_edges(self):
        if (
            self.byte_interval is None
            or self.byte_interval.section is None
            or self.byte_interval.section.module is None
            or self.byte_interval.section.module.ir is None
        ):
            return ()
        return (
            e
            for e in self.byte_interval.section.module.ir.cfg
            if e.target == self
        )

    @property
    def outgoing_edges(self):
        if (
            self.byte_interval is None
            or self.byte_interval.section is None
            or self.byte_interval.section.module is None
            or self.byte_interval.section.module.ir is None
        ):
            return ()
        return (
            e
            for e in self.byte_interval.section.module.ir.cfg
            if e.source == self
        )


class ProxyBlock(CfgNode):
    """A placeholder that serves as the endpoint (source or target) of a
    :class:`gtirb.Edge`.

    ProxyBlock objects allow the construction of CFG edges to or from
    another node. For example, a call to a function in another module
    may be represented by a :class:`gtirb.Edge` that originates at the
    calling :class:`gtirb.CodeBlock` and targets a ProxyBlock. Another
    example would be a :class:`gtirb.Edge` that represents an indirect
    jump whose target is not known.

    A ProxyBlock does not represent any instructions and so has neither
    an address nor a size.
    """

    def __init__(self, *, uuid=None):
        super().__init__(uuid=uuid)
        self._module = None  # type: "Module"

    @classmethod
    def _decode_protobuf(
        cls,
        proto_proxy,  # type: ProxyBlock_pb2.ProxyBlock
        uuid,  # type: UUID
        ir,  # type: typing.Optional["IR"]
    ):
        # type: (...) -> ProxyBlock
        b = cls(uuid=uuid)
        b._add_to_uuid_cache(ir._local_uuid_cache)
        return b

    def _to_protobuf(self):
        # type: () -> ProxyBlock_pb2.ProxyBlock
        proto_proxyblock = ProxyBlock_pb2.ProxyBlock()
        proto_proxyblock.uuid = self.uuid.bytes
        return proto_proxyblock

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, ProxyBlock):
            return False
        return self.uuid == other.uuid

    def __repr__(self):
        # type: () -> str
        return "ProxyBlock(" "uuid={uuid!r}, " ")".format(**self.__dict__)

    @property
    def module(self):
        # type: () -> "Module"
        return self._module

    @module.setter
    def module(self, value):
        # type: ("Module") -> None
        if self._module is not None:
            self._module.proxies.discard(self)
        if value is not None:
            value.proxies.add(self)

    @property
    def references(self):
        if self.module is None:
            return ()
        return (s for s in self.module.symbols if s.referent == self)

    @property
    def incoming_edges(self):
        if self.module is None or self.module.ir is None:
            return ()
        return (e for e in self.module.ir.cfg if e.target == self)

    @property
    def outgoing_edges(self):
        if self.module is None or self.module.ir is None:
            return ()
        return (e for e in self.module.ir.cfg if e.source == self)

    @property
    def ir(self):
        # type: () -> "IR"
        """Get the IR this node ultimately belongs to."""
        if self.module is None:
            return None
        return self.module.ir



"""The IR is the core class for reading and writing GTIRB files.

    You can open a GTIRB Protobuf file and load it into an IR instance:

    >>> ir = IR.load_protobuf('filename.gtirb')

    And then you can write the IR instance as a Protobuf file:

    >>> ir.save_protobuf('filename.gtirb')
"""

import itertools
import typing
from uuid import UUID

from .auxdata import AuxData, AuxDataContainer
from .block import ByteBlock, CfgNode, CodeBlock, DataBlock, ProxyBlock
from .byteinterval import ByteInterval, SymbolicExpressionElement
from .cfg import Edge
from .module import Module
from .node import Node
from .proto import CFG_pb2, IR_pb2
from .section import Section
from .symbol import Symbol
from .util import (
    DictLike,
    ListWrapper,
    nodes_at,
    nodes_on,
    symbolic_expressions_at,
)
from .version import PROTOBUF_VERSION


class IR(AuxDataContainer):
    """A complete internal representation consisting of multiple Modules.

    :ivar ~.modules: A list of :class:`Module`\\s contained in the IR.
    :ivar ~.cfg: A set of :class:`Edge`\\s representing the IR's control
        flow graph.
    :ivar ~.version: The Protobuf version of this IR.
    """

    class _ModuleList(ListWrapper[Module]):
        def __init__(self, node, *args):
            self._node = node  # type: IR
            super().__init__(*args)

        def _remove(self, v):
            v._ir = None
            v._remove_from_uuid_cache(self._node._local_uuid_cache)

        def _add(self, v):
            if v._ir is not None:
                v._ir.modules.remove(v)
            v._ir = self._node
            v._add_to_uuid_cache(self._node._local_uuid_cache)

        def __setitem__(self, i, v):
            self._remove(self[i])
            super().__setitem__(i, v)
            self._add(v)

        def __delitem__(self, i):
            self._remove(self[i])
            super().__delitem__(i)

        def insert(self, i, v):
            self._add(v)
            return super().insert(i, v)

    def __init__(
        self,
        *,
        modules=list(),  # type: typing.Iterable[Module]
        aux_data=dict(),  # type: DictLike[str, AuxData]
        cfg=set(),  # type: typing.Iterable[Edge]
        version=PROTOBUF_VERSION,  # type: int
        uuid=None  # type: typing.Optional[UUID]
    ):
        # type: (...) -> None
        """
        :param modules: A list of Modules contained in the IR.
        :param cfg: A set of :class:`Edge`\\s representing the IR's control
            flow graph. Defaults to being empty.
        :param aux_data: The initial auxiliary data to be associated
            with the object, as a mapping from names to
            :class:`gtirb.AuxData`. Defaults to being empty.
        :param version: The Protobuf version of this IR.
        :param uuid: The UUID of this ``IR``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        """

        self._local_uuid_cache = dict()  # type: typing.Dict[UUID, Node]
        # Modules are decoded before the aux data, since the UUID decoder
        # checks Node's cache.
        self.modules = IR._ModuleList(
            self, modules
        )  # type: typing.List[Module]
        self.cfg = set(cfg)  # type: typing.Set[Edge]
        self.version = version  # type: int
        super().__init__(aux_data, uuid)
        self._local_uuid_cache[self.uuid] = self

    @classmethod
    def _decode_protobuf(cls, proto_ir, uuid, _):
        # type: (IR_pb2.IR, UUID, typing.Optional["IR"]) -> IR
        if proto_ir.version != PROTOBUF_VERSION:
            raise ValueError(
                "Attempt to decode IR of version %s (expected version %s)"
                % (proto_ir.version, PROTOBUF_VERSION)
            )

        ir = cls(version=proto_ir.version, uuid=uuid)
        ir.modules.extend(
            Module._from_protobuf(m, ir) for m in proto_ir.modules
        )
        ir.cfg.update(
            Edge._from_protobuf(e, ir.get_by_uuid) for e in proto_ir.cfg.edges
        )
        ir.aux_data.update(
            AuxDataContainer._read_protobuf_aux_data(proto_ir, ir)
        )
        return ir

    def _to_protobuf(self):
        # type: () -> IR_pb2.IR
        proto_ir = IR_pb2.IR()
        proto_ir.uuid = self.uuid.bytes
        proto_ir.version = self.version
        proto_ir.modules.extend(m._to_protobuf() for m in self.modules)
        proto_cfg = CFG_pb2.CFG()
        proto_cfg.vertices.extend(v.uuid.bytes for v in self.cfg_nodes)
        proto_cfg.edges.extend(e._to_protobuf() for e in self.cfg)
        proto_ir.cfg.CopyFrom(proto_cfg)
        self._write_protobuf_aux_data(proto_ir)
        return proto_ir

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, IR) or not super().deep_eq(other):
            return False
        self_modules = sorted(self.modules, key=lambda m: m.uuid)
        other_modules = sorted(other.modules, key=lambda m: m.uuid)
        if not len(self_modules) == len(other_modules):
            return False
        for self_module, other_module in zip(self_modules, other_modules):
            if not self_module.deep_eq(other_module):
                return False
        self_edges = sorted(
            self.cfg, key=lambda e: (e.source.uuid, e.target.uuid)
        )
        other_edges = sorted(
            other.cfg, key=lambda e: (e.source.uuid, e.target.uuid)
        )
        if not len(self_edges) == len(other_edges):
            return False
        for self_edge, other_edge in zip(self_edges, other_edges):
            if not self_edge.deep_eq(other_edge):
                return False
        return self.version == other.version

    @staticmethod
    def load_protobuf_file(protobuf_file):
        # type: (typing.BinaryIO) -> IR
        """Load IR from a Protobuf object.

        Use this function when you have a Protobuf object already loaded,
        and you want to parse it as a GTIRB IR.
        If the Protobuf object is stored in a file,
        use :func:`gtirb.IR.load_protobuf` instead.

        :param protobuf_file: A byte stream encoding a GTIRB Protobuf message.
        :returns: An IR object representing the same
            information that is contained in ``protobuf_file``.
        """

        ir = IR_pb2.IR()
        ir.ParseFromString(protobuf_file.read())
        return IR._from_protobuf(ir, None)

    @staticmethod
    def load_protobuf(file_name):
        # type: (str) -> IR
        """Load IR from a Protobuf file at the specified path.

        :param file_name: The path to the Protobuf file.
        :returns: A Python GTIRB IR object.
        """
        with open(file_name, "rb") as f:
            return IR.load_protobuf_file(f)

    def save_protobuf_file(self, protobuf_file):
        # type: (typing.BinaryIO) -> None
        """Save ``self`` to a Protobuf object.

        :param protobuf_file: The byte stream to write the GTIRB Protobuf
            message to.
        """

        protobuf_file.write(self._to_protobuf().SerializeToString())

    def save_protobuf(self, file_name):
        # type: (str) -> None
        """Save ``self`` to a Protobuf file at the specified path.

        :param file_name: The file path at which to
            save the Protobuf representation of ``self``.
        """
        with open(file_name, "wb") as f:
            self.save_protobuf_file(f)

    def __repr__(self):
        # type: () -> str
        return (
            "IR("
            "uuid={uuid!r}, "
            "modules={modules!r}, "
            "cfg={cfg!r}, "
            "version={version}, "
            ")".format(**self.__dict__)
        )

    @property
    def proxy_blocks(self):
        # type: () -> typing.Iterator[ProxyBlock]
        """The :class:`ProxyBlock`\\s in this IR."""

        return itertools.chain.from_iterable(m.proxies for m in self.modules)

    @property
    def sections(self):
        # type: () -> typing.Iterator[Section]
        """The :class:`Section`\\s in this IR."""

        return itertools.chain.from_iterable(m.sections for m in self.modules)

    @property
    def symbols(self):
        # type: () -> typing.Iterator[Symbol]
        """The :class:`Symbol`\\s in this IR."""

        return itertools.chain.from_iterable(m.symbols for m in self.modules)

    @property
    def byte_intervals(self):
        # type: () -> typing.Iterator[ByteInterval]
        """The :class:`ByteInterval`\\s in this IR."""

        return itertools.chain.from_iterable(
            m.byte_intervals for m in self.modules
        )

    @property
    def byte_blocks(self):
        # type: () -> typing.Iterator[ByteBlock]
        """The :class:`ByteBlock`\\s in this IR."""

        return itertools.chain.from_iterable(
            m.byte_blocks for m in self.modules
        )

    @property
    def code_blocks(self):
        # type: () -> typing.Iterator[CodeBlock]
        """The :class:`CodeBlock`\\s in this IR."""

        return itertools.chain.from_iterable(
            m.code_blocks for m in self.modules
        )

    @property
    def data_blocks(self):
        # type: () -> typing.Iterator[DataBlock]
        """The :class:`DataBlock`\\s in this IR."""

        return itertools.chain.from_iterable(
            m.data_blocks for m in self.modules
        )

    @property
    def cfg_nodes(self):
        # type: () -> typing.Iterator[CfgNode]
        """The :class:`CfgNode`\\s in this IR."""

        return itertools.chain.from_iterable(m.cfg_nodes for m in self.modules)

    def modules_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[Module]
        """Finds all the modules that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.modules, addrs)

    def modules_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[Module]
        """Finds all the modules that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.modules, addrs)

    def sections_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[Section]
        """Finds all the sections that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.sections, addrs)

    def sections_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[Section]
        """Finds all the sections that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.sections, addrs)

    def byte_intervals_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteInterval]
        """Finds all the byte intervals that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.byte_intervals, addrs)

    def byte_intervals_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteInterval]
        """Finds all the byte intervals that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.byte_intervals, addrs)

    def byte_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
        """Finds all the byte blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.byte_blocks, addrs)

    def byte_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
        """Finds all the byte blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.byte_blocks, addrs)

    def code_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
        """Finds all the code blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.code_blocks, addrs)

    def code_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
        """Finds all the code blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.code_blocks, addrs)

    def data_blocks_on(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
        """Finds all the data blocks that overlap an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_on(self.data_blocks, addrs)

    def data_blocks_at(self, addrs):
        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
        """Finds all the data blocks that begin at an address or range of
        addresses.

        :param addrs: Either a ``range`` object or a single address.
        """

        return nodes_at(self.data_blocks, addrs)

    def symbolic_expressions_at(
        self, addrs  # type: typing.Union[int, range]
    ):
        # type: (...) -> typing.Iterable[SymbolicExpressionElement]
        """Finds all the symbolic expressions that begin at an address or
        range of addresses.

        :param addrs: Either a ``range`` object or a single address.
        :returns: Yields ``(interval, offset, symexpr)`` tuples for every
            symbolic expression in the range.
        """

        return symbolic_expressions_at(self.modules, addrs)

    def get_by_uuid(self, uuid):
        # type: (UUID) -> typing.Optional[Node]
        """Look up a node by its UUID.

        This method will find any node currently attached to this IR.
        It will not find any nodes attached to other IRs, or not attached to
        any IR.

        :param uuid: The UUID to look up.
        :returns: The Node this UUID corresponds to, or None if no node exists
            with that UUID.
        """

        return self._local_uuid_cache.get(uuid)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gtirb/proto/Offset.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='gtirb/proto/Offset.proto',
  package='gtirb.proto',
  syntax='proto3',
  serialized_pb=_b('\n\x18gtirb/proto/Offset.proto\x12\x0bgtirb.proto\"2\n\x06Offset\x12\x12\n\nelement_id\x18\x01 \x01(\x0c\x12\x14\n\x0c\x64isplacement\x18\x02 \x01(\x04\x42\x1c\n\x1a\x63om.grammatech.gtirb.protob\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_OFFSET = _descriptor.Descriptor(
  name='Offset',
  full_name='gtirb.proto.Offset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='element_id', full_name='gtirb.proto.Offset.element_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='displacement', full_name='gtirb.proto.Offset.displacement', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41,
  serialized_end=91,
)

DESCRIPTOR.message_types_by_name['Offset'] = _OFFSET

Offset = _reflection.GeneratedProtocolMessageType('Offset', (_message.Message,), dict(
  DESCRIPTOR = _OFFSET,
  __module__ = 'gtirb.proto.Offset_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.Offset)
  ))
_sym_db.RegisterMessage(Offset)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\032com.grammatech.gtirb.proto'))
# @@protoc_insertion_point(module_scope)



import typing
from uuid import UUID

from .block import Block
from .node import Node
from .proto import Symbol_pb2

Payload = typing.Union[Block, int]
"""A type hint representing the possible Symbol payloads."""


class Symbol(Node):
    """Represents a symbol, which maps a name to an object in the IR.

    :ivar ~.name: The name of this symbol.
    :ivar ~.at_end: True if this symbol is at the end of its referent, rather
        than at the beginning. Has no meaning for integral symbols.
    """

    def __init__(
        self,
        name,  # type: str
        uuid=None,  # type: typing.Optional[UUID]
        payload=None,  # type: typing.Optional[Payload]
        at_end=False,  # type: bool
    ):
        """
        :param name: The name of this symbol.
        :param uuid: The UUID of this ``Symbol``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        :param payload: The value this symbol points to.
            May be an address, a Node, or None.
        :param at_end: True if this symbol is at the end of its referent,
            rather than at the beginning.
        """

        super().__init__(uuid)
        self.name = name  # type: str
        self.at_end = at_end  # type: bool
        self._payload = payload  # type: typing.Optional[Payload]
        self._module = None  # type: "Module"

    @property
    def value(self):
        # type: () -> typing.Optional[int]
        """The value of a Symbol, which is an integer or None.
        ``value`` and ``referent`` are mutually exclusive.
        """

        if not isinstance(self._payload, Block):
            return self._payload
        return None

    @property
    def referent(self):
        # type: () -> typing.Optional[Block]
        """The object referred to by a Symbol, which is :class:`Block`
        or None. ``value`` and ``referent`` are mutually exclusive.
        """

        if isinstance(self._payload, Block):
            return self._payload
        return None

    @value.setter
    def value(self, value):
        # type: (typing.Optional[int]) -> None
        self._payload = value

    @referent.setter
    def referent(self, referent):
        # type: (typing.Optional[Block]) -> None
        self._payload = referent

    @classmethod
    def _decode_protobuf(cls, proto_symbol, uuid, ir):
        # type: (Symbol_pb2.Symbol,UUID, typing.Optional["IR"]) -> Symbol
        symbol = cls(
            name=proto_symbol.name, at_end=proto_symbol.at_end, uuid=uuid
        )
        if proto_symbol.HasField("value"):
            symbol.value = proto_symbol.value
        if proto_symbol.HasField("referent_uuid"):
            referent_uuid = UUID(bytes=proto_symbol.referent_uuid)
            try:
                symbol.referent = ir.get_by_uuid(referent_uuid)
            except KeyError as e:
                raise KeyError("Could not find referent UUID %s" % e)
        symbol._add_to_uuid_cache(ir._local_uuid_cache)
        return symbol

    def _to_protobuf(self):
        # type: () -> Symbol_pb2.Symbol
        proto_symbol = Symbol_pb2.Symbol()
        proto_symbol.uuid = self.uuid.bytes
        if self.value is not None:
            proto_symbol.value = self.value
        elif self.referent is not None:
            proto_symbol.referent_uuid = self.referent.uuid.bytes
        proto_symbol.name = self.name
        proto_symbol.at_end = self.at_end
        return proto_symbol

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, Symbol):
            return False
        if self.value != other.value:
            return False
        if self.referent is None:
            if other.referent is not None:
                return False
        else:
            if not self.referent.deep_eq(other.referent):
                return False
        return (
            self.name == other.name
            and self.at_end == other.at_end
            and self.uuid == other.uuid
        )

    def __repr__(self):
        # type: () -> str
        return (
            "Symbol("
            "uuid={uuid!r}, "
            "name={name!r}, "
            "payload={_payload!r}, "
            "at_end={at_end!r}, "
            ")".format(**self.__dict__)
        )

    @property
    def module(self):
        # type: () -> "Module"
        return self._module

    @module.setter
    def module(self, value):
        # type: ("Module") -> None
        if self._module is not None:
            self._module.symbols.discard(self)
        if value is not None:
            value.symbols.add(self)

    def _add_to_uuid_cache(self, cache):
        # type: (typing.Dict[UUID, Node]) -> None
        """Update the UUID cache when this node is added."""

        cache[self.uuid] = self

    def _remove_from_uuid_cache(self, cache):
        # type: (typing.Dict[UUID, Node]) -> None
        """Update the UUID cache when this node is removed."""

        del cache[self.uuid]

    @property
    def ir(self):
        # type: () -> "IR"
        """Get the IR this node ultimately belongs to."""
        if self.module is None:
            return None
        return self.module.ir



import typing
from enum import Enum
from uuid import UUID

from .block import CfgNode
from .node import Node
from .proto import CFG_pb2


class Edge:
    """An edge in the CFG from ``source`` to ``target``, with optional
    control-flow details in ``label``.

    :ivar ~.source: The source CFG node.
    :ivar ~.target: The target CFG node.
    :ivar ~.label: An optional label containing more control flow information.
    """

    class Type(Enum):
        """The type of control flow transfer indicated by a
        :class:`gtirb.Edge`.
        """

        Branch = CFG_pb2.EdgeType.Value("Type_Branch")
        """This edge is the explicit target of a jump instruction.
        May be conditional or unconditional. If conditional, there will be
        a corresponding edge of type :attr:`gtirb.Edge.Type.Fallthrough`.
        """

        Call = CFG_pb2.EdgeType.Value("Type_Call")
        """This edge is the explicit target of a call instruction.
        Unless the function does not return, there will also be a
        corresponding edge of type :attr:`gtirb.Edge.Type.Fallthrough`.
        """

        Fallthrough = CFG_pb2.EdgeType.Value("Type_Fallthrough")
        """This edge represents two blocks executing in sequence.
        This occurs on the non-branching paths of conditional branch
        instructions, after call instructons have returned, and when two
        blocks have no control flow between them, but another
        :class:`gtirb.Edge` targets the target block.
        If there exists a fallthrough edge from block ``A`` to block ``B``,
        then ``A`` must immediately precede ``B`` in memory.
        """

        Return = CFG_pb2.EdgeType.Value("Type_Return")
        """This edge represents a return from a function, generally via a
        return instruction. Return edges may be omitted from valid CFGs;
        a function may have an uncomputable number of possible return sites,
        due to the possibility of indirect calls.
        """

        Syscall = CFG_pb2.EdgeType.Value("Type_Syscall")
        """This edge is the explicit target of a system call instruction.
        Unless the function does not return, there will also be a
        corresponding edge of type :attr:`gtirb.Edge.Type.Fallthrough`. This
        is the system call equivalent to :class:`gtirb.Edge.Type.Call`.
        """

        Sysret = CFG_pb2.EdgeType.Value("Type_Sysret")
        """This edge represents a return from a system call, generally via a
        return instruction. Return edges may be omitted from valid CFGs;
        a function may have an uncomputable number of possible return sites,
        due to the possibility of indirect calls. This is the system call
        equivalent to :class:`gtirb.Edge.Type.Return`.
        """

    class Label:
        """Contains a more detailed description of a :class:`gtirb.Edge`
        in the CFG.

        :ivar ~.conditional: When this edge is part of a conditional branch,
            ``conditional`` is ``True`` when the edge represents the control
            flow taken when the branch's condition is met, and ``False``
            when it represents the control flow taken when the branch's
            condition is not met. Otherwise, it is always ``False``.
        :ivar ~.direct: ``True`` if the branch or call is direct,
                and ``False`` if it is indirect. If an edge is indirect,
                then all outgoing indirect edges represent the set of
                possible locations the edge may branch to. If there
                exists an indirect outgoing edge to a :class:`gtirb.ProxyBlock`
                without any :class:`gtirb.Symbol` objects referring to it,
                then the set of all possible branch locations is unknown.
        :ivar ~.type: The type of control flow the :class:`gtirb.Edge`
            represents.
        """

        def __init__(
            self,
            type,  # type: Edge.Type
            *,
            conditional=False,  # type: bool
            direct=True  # type: bool
        ):
            """
            :param type: The type of control flow the :class:`gtirb.Edge`
                represents.
            :param conditional: ``True`` when this edge represents a
                conditional branch taken when the branch's condition is met,
                and ``False`` otherwise.
            :param direct: ``True`` if the branch or call is direct,
                and ``False`` if it is indirect.
            """

            self.type = type  # type: Edge.Type
            self.conditional = conditional  # type: bool
            self.direct = direct  # type: bool

        @classmethod
        def _from_protobuf(cls, label):
            # type: (CFG_pb2.EdgeLabel) -> Edge.Label
            return Edge.Label(
                type=Edge.Type(label.type),
                conditional=label.conditional,
                direct=label.direct,
            )

        def _to_protobuf(self):
            # type: () -> CFG_pb2.EdgeLabel
            proto_label = CFG_pb2.EdgeLabel()
            proto_label.type = self.type.value
            proto_label.conditional = self.conditional
            proto_label.direct = self.direct
            return proto_label

        def __eq__(self, other):
            # type: (typing.Any) -> bool
            if not isinstance(other, Edge.Label):
                return False
            return (
                self.type == other.type
                and self.conditional == other.conditional
                and self.direct == other.direct
            )

        def __hash__(self):
            # type: () -> int
            return hash((self.type, self.conditional, self.direct))

        def __repr__(self):
            # type: () -> str
            return (
                "Edge.Label("
                "type=Edge.{type!s}, "
                "conditional={conditional!r}, "
                "direct={direct!r}, "
                ")".format(**self.__dict__)
            )

    def __init__(
        self,
        source,  # type: CfgNode
        target,  # type: CfgNode
        label=None,  # type: typing.Optional["Edge.Label"]
    ):
        # type: (...) -> None
        """
        :param source: The source CFG node.
        :param target: The target CFG node.
        :param label: An optional label
            containing more control flow information.
        """
        self.source = source  # type: CfgNode
        self.target = target  # type: CfgNode
        self.label = label  # type: typing.Optional["Edge.Label"]

    @classmethod
    def _from_protobuf(cls, edge, get_by_uuid):
        # type: (CFG_pb2.Edge, typing.Callable[[UUID], Node]) -> Edge
        source_uuid = UUID(bytes=edge.source_uuid)
        target_uuid = UUID(bytes=edge.target_uuid)

        source = get_by_uuid(source_uuid)
        target = get_by_uuid(target_uuid)
        if not isinstance(source, CfgNode):
            raise ValueError(
                "In CFG: source UUID %s is a %s, not a CfgNode"
                % (source_uuid, type(source).__name__)
            )
        if not isinstance(target, CfgNode):
            raise ValueError(
                "In CFG: target UUID %s is a %s, not a CfgNode"
                % (target_uuid, type(target).__name__)
            )

        label = None
        if edge.label is not None:
            label = Edge.Label._from_protobuf(edge.label)
        return cls(source, target, label)

    def _to_protobuf(self):
        # type: () -> CFG_pb2.Edge
        proto_edge = CFG_pb2.Edge()
        proto_edge.source_uuid = self.source.uuid.bytes
        proto_edge.target_uuid = self.target.uuid.bytes
        if self.label is not None:
            proto_edge.label.CopyFrom(self.label._to_protobuf())
        return proto_edge

    def __eq__(self, other):
        # type: (typing.Any) -> bool
        if not isinstance(other, Edge):
            return False
        return (
            self.source.uuid == other.source.uuid
            and self.target.uuid == other.target.uuid
            and self.label == other.label
        )

    def deep_eq(self, other):
        # type: (typing.Any) -> bool
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, Edge):
            return False
        return (
            self.source.deep_eq(other.source)
            and self.target.deep_eq(other.target)
            and self.label == other.label
        )

    def __hash__(self):
        # type: () -> int
        return hash((self.source.uuid, self.target.uuid, self.label))

    def __repr__(self):
        # type: () -> str
        return (
            "Edge("
            "source={source!r}, "
            "target={target!r}, "
            "label={label!r}, "
            ")".format(**self.__dict__)
        )



__all__ = [
    "AuxData",
    "AuxDataContainer",
    "Block",
    "ByteBlock",
    "ByteInterval",
    "CfgNode",
    "CodeBlock",
    "DataBlock",
    "Edge",
    "IR",
    "Module",
    "Node",
    "Offset",
    "ProxyBlock",
    "Section",
    "Serialization",
    "Symbol",
    "SymAddrAddr",
    "SymAddrConst",
    "SymStackConst",
]

from .auxdata import AuxData, AuxDataContainer
from .block import Block, ByteBlock, CfgNode, CodeBlock, DataBlock, ProxyBlock
from .byteinterval import ByteInterval
from .cfg import Edge
from .ir import IR
from .module import Module
from .node import Node
from .offset import Offset
from .section import Section
from .serialization import Serialization
from .symbol import Symbol
from .symbolicexpression import SymAddrAddr, SymAddrConst, SymStackConst
from .version import API_VERSION

__version__ = API_VERSION



API_VERSION = (
    "1."
    "8."
    "6"
)  # type: str
"""The semantic version of this API."""

PROTOBUF_VERSION = 1  # type: int
"""The version of Protobuf this API can read and write from.
Attempts to load old Protobuf versions will raise a ``ValueError``.
"""



import io
import typing
from re import findall
from uuid import UUID

from .node import Node
from .offset import Offset

if typing.TYPE_CHECKING:
    CacheLookupFn = typing.Optional[
        typing.Callable[[UUID], typing.Optional[Node]]
    ]


class CodecError(Exception):
    """Base class for codec exceptions."""


class DecodeError(CodecError):
    """An exception during decoding."""


class EncodeError(CodecError):
    """An exception during encoding."""


class TypeNameError(EncodeError):
    """A type name is malformed."""

    def __init__(self, hint):
        # type: (str) -> None
        super().__init__("malformed type name: '%s'" % (hint))


class UnknownCodecError(CodecError):
    """An unknown codec name is encountered.
    Caught and handled by the top-level codec methods.

    :param name: the name of the unknown codec
    """

    def __init__(self, name):
        # type: (str) -> None
        self.name = name


SubtypeTree = typing.Tuple[str, typing.Iterable["SubtypeTree"]]
"""A type hint representing a parsed serialization type name.
A ``SubtypeTree`` is a ``tuple`` with two items: A ``str`` giving
the name of the type and a ``tuple`` of type parameters
(which are also ``SubtypeTree``\\s). For example, the following are all
valid ``SubtypeTree``\\s:

>>> ('string', ())
>>> ('sequence', (('UUID',()),))
>>> ('mapping', (('string', ()),('set', (('UUID', ()),))))
"""


class Codec:
    """The base class for codecs."""

    @staticmethod
    def decode(
        raw_bytes,  # type: typing.BinaryIO
        *,
        serialization=None,  # type: Serialization
        subtypes=tuple(),  # type: SubtypeTree
        get_by_uuid=None  # type: CacheLookupFn
    ):
        # type: (...) -> typing.Any
        """Decode the specified raw data into a Python object.

        :param raw_bytes: The BytesIO object to be decoded.
        :param serialization: A Serialization instance used to invoke
            other codecs if needed.
        :param subtypes: The parsed type of this object.
        :param get_by_uuid: A function to look up nodes by UUID.
        :returns: A new Python object, as decoded from ``raw_bytes``.
        """

        raise NotImplementedError

    @staticmethod
    def encode(
        out,  # type: typing.BinaryIO
        item,  # type: typing.Any
        *,
        serialization=None,  # type: Serialization
        subtypes=tuple()  # type: SubtypeTree
    ):
        # type: (...) -> None
        """Encode an item, writing the serialized object to ``out``.

        :param out: A binary stream to serialize to.
        :param item: The arbitrary Python object to encode.
        :param serialization: A Serialization instance, used to invoke
            other codecs if needed.
        :param subtypes: The parsed type of this object.
        """

        raise NotImplementedError


class Int64Codec(Codec):
    """A Codec for 64-bit signed integers."""

    @staticmethod
    def decode(
        raw_bytes, *, serialization=None, subtypes=tuple(), get_by_uuid=None
    ):
        if subtypes != ():
            raise DecodeError("int64_t should have no subtypes")
        return int.from_bytes(
            raw_bytes.read(8), byteorder="little", signed=True
        )

    @staticmethod
    def encode(out, val, *, serialization=None, subtypes=tuple()):
        if subtypes != ():
            raise EncodeError("int64_t should have no subtypes")
        out.write(val.to_bytes(8, byteorder="little", signed=True))


class MappingCodec(Codec):
    """A Codec for mapping<K,V> entries. Implemented via ``dict``."""

    @staticmethod
    def decode(raw_bytes, *, serialization, subtypes, get_by_uuid=None):
        try:
            key_type, val_type = subtypes
        except (TypeError, ValueError):
            raise DecodeError(
                "could not unpack mapping types: %s" % str(subtypes)
            )
        mapping = dict()
        mapping_len = Uint64Codec.decode(raw_bytes)
        for _ in range(mapping_len):
            key = serialization._decode_tree(raw_bytes, key_type, get_by_uuid)
            val = serialization._decode_tree(raw_bytes, val_type, get_by_uuid)
            mapping[key] = val
        return mapping

    @staticmethod
    def encode(out, mapping, *, serialization, subtypes):
        try:
            key_type, val_type = subtypes
        except (TypeError, ValueError):
            raise EncodeError(
                "could not unpack mapping types: %s" % str(subtypes)
            )
        Uint64Codec.encode(out, len(mapping))
        for key, val in mapping.items():
            serialization._encode_tree(out, key, key_type)
            serialization._encode_tree(out, val, val_type)


class OffsetCodec(Codec):
    """A Codec for :class:`gtirb.Offset` objects,
    containing a UUID and a displacement.
    """

    @staticmethod
    def decode(
        raw_bytes, *, serialization=None, subtypes=tuple(), get_by_uuid=None
    ):
        if subtypes != ():
            raise DecodeError("Offset should have no subtypes")
        element_uuid = UUIDCodec.decode(raw_bytes, get_by_uuid=get_by_uuid)
        displacement = Uint64Codec.decode(raw_bytes)

        return Offset(element_uuid, displacement)

    @staticmethod
    def encode(out, val, *, serialization=None, subtypes=tuple()):
        if subtypes != ():
            raise EncodeError("Offset should have no subtypes")
        UUIDCodec.encode(out, val.element_id)
        Uint64Codec.encode(out, val.displacement)


class SequenceCodec(Codec):
    """A Codec for sequence<T> entries. Implemented via ``list``."""

    @staticmethod
    def decode(raw_bytes, *, serialization, subtypes, get_by_uuid=None):
        try:
            (subtype,) = subtypes
        except (TypeError, ValueError) as e:
            raise DecodeError("could not unpack sequence type: %s" % str(e))
        sequence = list()
        sequence_len = Uint64Codec.decode(raw_bytes)
        for _ in range(sequence_len):
            sequence.append(
                serialization._decode_tree(raw_bytes, subtype, get_by_uuid)
            )
        return sequence

    @staticmethod
    def encode(out, sequence, *, serialization, subtypes):
        try:
            (subtype,) = subtypes
        except (TypeError, ValueError) as e:
            raise EncodeError("could not unpack sequence type: %s" % str(e))
        Uint64Codec.encode(out, len(sequence))
        for item in sequence:
            serialization._encode_tree(out, item, subtype)


class SetCodec(Codec):
    """A Codec for set<T> entries. Implemented via ``set``."""

    @staticmethod
    def decode(raw_bytes, *, serialization, subtypes, get_by_uuid=None):
        try:
            (subtype,) = subtypes
        except (TypeError, ValueError) as e:
            raise DecodeError("could not unpack set type: %s" % str(e))
        decoded_set = set()
        set_len = Uint64Codec.decode(raw_bytes)
        for _ in range(set_len):
            decoded_set.add(
                serialization._decode_tree(raw_bytes, subtype, get_by_uuid)
            )
        return decoded_set

    @staticmethod
    def encode(out, items, *, serialization, subtypes):
        try:
            (subtype,) = subtypes
        except (TypeError, ValueError) as e:
            raise EncodeError("could not unpack set type: %s" % str(e))
        Uint64Codec.encode(out, len(items))
        for item in items:
            serialization._encode_tree(out, item, subtype)


class TupleCodec(Codec):
    """A Codec for tuple<...> entries. Implemented via ``tuple``."""

    @staticmethod
    def decode(raw_bytes, *, serialization, subtypes, get_by_uuid=None):
        # The length of a tuple is not contained in the Protobuf
        # representation, so error checking cannot be done here.
        decoded_list = list()
        for subtype in subtypes:
            decoded_list.append(
                serialization._decode_tree(raw_bytes, subtype, get_by_uuid)
            )
        return tuple(decoded_list)

    @staticmethod
    def encode(out, items, *, serialization, subtypes):
        if len(items) != len(subtypes):
            raise EncodeError("length of tuple does not match subtype count")
        for item, subtype in zip(items, subtypes):
            serialization._encode_tree(out, item, subtype)


class StringCodec(Codec):
    """A Codec for strings."""

    @staticmethod
    def decode(
        raw_bytes, *, serialization=None, subtypes=tuple(), get_by_uuid=None
    ):
        if subtypes != tuple():
            raise DecodeError("string should have no subtypes")
        size = Uint64Codec.decode(raw_bytes)
        return str(raw_bytes.read(size), "utf-8")

    @staticmethod
    def encode(out, val, *, serialization=None, subtypes=tuple()):
        if subtypes != ():
            raise EncodeError("string should have no subtypes")
        Uint64Codec.encode(out, len(val))
        out.write(val.encode())


class Uint64Codec(Codec):
    """A Codec for 64-bit unsigned integers."""

    @staticmethod
    def decode(
        raw_bytes, *, serialization=None, subtypes=tuple(), get_by_uuid=None
    ):
        if subtypes != ():
            raise DecodeError("uint64_t should have no subtypes")
        return int.from_bytes(
            raw_bytes.read(8), byteorder="little", signed=False
        )

    @staticmethod
    def encode(out, val, *, serialization=None, subtypes=tuple()):
        if subtypes != ():
            raise EncodeError("uint64_t should have no subtypes")
        out.write(val.to_bytes(8, byteorder="little"))


class UUIDCodec(Codec):
    """A Codec for raw UUIDs or Nodes.

    Decoding a UUID first checks the Node cache for an object with the
    corresponding UUID, and either returns the object it hits or a new
    raw UUID.
    """

    @staticmethod
    def decode(
        raw_bytes, *, serialization=None, subtypes=tuple(), get_by_uuid=None
    ):
        if subtypes != ():
            raise DecodeError("UUID should have no subtypes")
        uuid = UUID(bytes=raw_bytes.read(16))
        existing_node = get_by_uuid(uuid)
        return uuid if existing_node is None else existing_node

    @staticmethod
    def encode(out, val, *, serialization=None, subtypes=tuple()):
        if subtypes != ():
            raise EncodeError("UUID should have no subtypes")
        if isinstance(val, Node):
            out.write(val.uuid.bytes)
        elif isinstance(val, UUID):
            out.write(val.bytes)
        else:
            raise EncodeError("UUID codec only supports UUIDs or Nodes")


class UnknownData(bytes):
    """This class is a blob of bytes representing data with an unknown type.
    Generated by :func:`gtirb.Serialization.decode` when it encounters
    the name of an unknown codec. Use only at the top level of an auxdata.
    """


class Serialization:
    """Manages codecs used to serialize and deserialize GTIRB objects.

    The :meth:`gtirb.Serialization.decode` method of
    :attr:`gtirb.AuxData.serializer` is called when GTIRB AuxData is loaded via
    :meth:`gtirb.IR.load_protobuf`, and the :meth:`gtirb.Serialization.encode`
    method of :attr:`gtirb.AuxData.serializer` is called when GTIRB AuxData is
    saved to file via :meth:`gtirb.IR.save_protobuf`. You can alter the
    encoding and decoding of AuxData values via
    :attr:`gtirb.Serialization.codecs`. To do this, create a new subclass of
    :class:`gtirb.serialization.Codec` and add it to
    :attr:`gtirb.Serialization.codecs`:

    >>> gtirb.AuxData.serializer.codecs['my_custom_type'] = MyCustomCodec

    This example registers a new type name, ``my_custom_type``, and associate
    it with a new codec, ``MyCustomCodec``.

    :ivar ~.codecs: A mapping of type names to codecs. Codecs can be added
        or overridden using this dictionary.
    """

    def __init__(self):
        """Initialize with the built-in `gtirb.serialization.Codec` subclasses.
        """

        self.codecs = {
            "Addr": Uint64Codec,
            "Offset": OffsetCodec,
            "int64_t": Int64Codec,
            "mapping": MappingCodec,
            "sequence": SequenceCodec,
            "set": SetCodec,
            "string": StringCodec,
            "tuple": TupleCodec,
            "uint64_t": Uint64Codec,
            "UUID": UUIDCodec,
        }  # type: typing.Dict[str, Codec]

    def _decode_tree(self, raw_bytes, type_tree, get_by_uuid):
        # type: (typing.BinaryIO, SubtypeTree, CacheLookupFn) -> typing.Any
        """Decode the data in ``raw_bytes`` given a parsed type tree.

        :param raw_bytes: The binary stream to read bytes from.
        :param type_tree: The parsed type of the object encoded by
            ``raw_bytes``.
        """

        try:
            type_name, subtypes = type_tree
        except ValueError:
            raise DecodeError("could not unpack type tree %s" % str(type_tree))
        if type_name not in self.codecs:
            raise UnknownCodecError(type_name)
        codec = self.codecs[type_name]
        return codec.decode(
            raw_bytes,
            serialization=self,
            subtypes=subtypes,
            get_by_uuid=get_by_uuid,
        )

    def _encode_tree(self, out, val, type_tree):
        # type: (typing.BinaryIO, typing.Any, SubtypeTree) -> None
        """Encode the data in ``val`` given a parsed type tree.

        :param out: A binary stream to write bytes to.
        :param val: The :class:`gtirb.AuxData` to encode.
        :param type_tree: The parsed type to encode ``val`` as.
        """

        try:
            type_name, subtypes = type_tree
        except ValueError:
            raise EncodeError("could not unpack type tree %s" % str(type_tree))
        if type_name not in self.codecs:
            raise UnknownCodecError(type_name)
        codec = self.codecs[type_name]
        return codec.encode(out, val, serialization=self, subtypes=subtypes)

    @staticmethod
    def _parse_type(type_name):
        # type: (str) -> SubtypeTree
        """Given an encoded aux_data type_name, generate its parse tree.

        >>> _parse_type('foo')
        ('foo', ())

        >>> _parse_type('foo<bar>')
        ('foo', (('bar',()),))

        >>> _parse_type('foo<bar<baz>>')
        ('foo', (('bar', (('baz', ()),)),))

        :param type_name: The type name to parse into a ``SubtypeTree``.
        """
        tokens = findall("[^<>,]+|<|>|,", type_name)

        def parse(tokens, tree):
            tree = list(tree)
            # It is an error to parse nothing
            if len(tokens) == 0:
                raise TypeNameError(type_name)
            first_token, *tail = tokens

            # The first token should be a name
            if first_token in {"<", ">", ","}:
                raise TypeNameError(type_name)

            # Base case
            if len(tail) == 0:
                tree.append((first_token, ()))
                return tuple(tree), []
            next_token, *tail = tail

            # No subtypes
            if next_token == ",":
                tree.append((first_token, ()))

            # Parse subtypes
            if next_token == "<":
                # Extract just the subtype tokens and parse them
                stack = ["<"]
                subtype_tokens = list()
                remaining_tokens = list()
                for t in tail:
                    if len(stack) == 0:
                        remaining_tokens.append(t)
                        continue
                    if t == "<":
                        stack.append(t)
                    elif t == ">":
                        stack.pop()
                    subtype_tokens.append(t)
                if len(stack) > 0 or subtype_tokens[-1] != ">":
                    raise TypeNameError(type_name)
                subtypes, remaining = parse(subtype_tokens[:-1], [])
                # Parsing should consume all subtype tokens
                if len(remaining) != 0:
                    raise TypeNameError(type_name)
                tree.append((first_token, subtypes))
                # Finish if all tokens are consumed
                if len(remaining_tokens) == 0:
                    return tuple(tree), []
                next_token, *tail = remaining_tokens

            # If the next token is a comma, parse next
            if next_token == ",":
                return parse(tail, tree)

            # None of the rules match, error
            raise TypeNameError(type_name)

        # There should only be one item at the root of the tree
        try:
            (parse_tree,) = parse(tokens, [])[0]
        except ValueError:
            raise TypeNameError(type_name)
        return parse_tree

    def decode(
        self,
        raw_bytes,  # type: typing.BinaryIO
        type_name,  # type: str
        get_by_uuid=None,  # type: CacheLookupFn
    ):
        # type: (...) -> typing.Any
        """Decode a :class:`gtirb.AuxData` of the specified type
        from the specified byte stream.

        :param raw_bytes: The byte stream from which to read the encoded value.
        :param type_name: The type name of the object encoded by ``raw_bytes``.
        :param get_by_uuid: A function to look up nodes by UUID.
        :returns: The object encoded by ``raw_bytes``.
        """

        parse_tree = Serialization._parse_type(type_name)
        all_bytes = None
        if isinstance(raw_bytes, (bytes, bytearray, memoryview)):
            all_bytes = raw_bytes
        else:
            all_bytes = raw_bytes.read()
        try:
            return self._decode_tree(
                io.BytesIO(all_bytes), parse_tree, get_by_uuid
            )
        except UnknownCodecError:
            # we found an unknwon codec; the entire data structure can't be
            # parsed; return a blob of bytes
            return UnknownData(all_bytes)

    def encode(self, out, val, type_name):
        # type: (typing.BinaryIO,typing.Any,str) -> None
        """Encodes the value of an AuxData value to bytes.

        :param out: A binary stream to write bytes to.
        :param val: The :class:`gtirb.AuxData` to encode.
        :param type_name: The type name of the value encapsulated
            by the :class:`gtirb.AuxData`.
        """

        if isinstance(val, UnknownData):
            # it was a blob of bytes because of a decoding problem;
            # just write the whole thing out
            out.write(val)
            return
        parse_tree = Serialization._parse_type(type_name)
        try:
            self._encode_tree(out, val, parse_tree)
        except UnknownCodecError as e:
            # rethrow UnknownCodecError, because we were supposed to catch it
            # via UnknownData. This means the user manually wrote a bad type.
            raise EncodeError("unknown codec: %s" % e.name)



import glob
import setuptools
import sys
import unittest
import os
import shutil

# if version of setuptools is 20.8.1 or below, it doesn't support env specs
if int(setuptools.__version__.split(".")[0]) < 21:
    if sys.version_info[:3] < (3, 5, 0):
        typing_dep = ["typing"]
    else:
        typing_dep = []
else:
    typing_dep = ["typing ; python_version<'3.5.0'"]


def gtirb_test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover("tests", pattern="test_*.py")
    return test_suite


# copy over files needed for sdist: README and LICENCE
root_dir = "/gt/gtirb"
this_dir = os.path.dirname(__file__)
shutil.copyfile(os.path.join(root_dir, "README.md"), "README")
shutil.copyfile(os.path.join(root_dir, "LICENSE.txt"), "LICENSE")

# run setuptools
if __name__ == "__main__":
    with open("README", "r") as fh:
        long_description = fh.read().replace(
            ".gtirb.svg",
            "https://raw.githubusercontent.com/"
            "GrammaTech/gtirb/master/.gtirb.svg",
        )

    setuptools.setup(
        name="gtirb",
        version=(
            "1"
            ".8"
            ".6"
        ),
        author="GrammaTech",
        author_email="gtirb@grammatech.com",
        description="GrammaTech Intermediate Representation for Binaries",
        packages=setuptools.find_packages(),
        test_suite="setup.gtirb_test_suite",
        install_requires=["protobuf<3.12.2"] + typing_dep,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        python_requires=">=3",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/grammatech/gtirb",
        license="MIT",
    )



from typing import NamedTuple
from uuid import UUID

from .node import Node
from .proto import Offset_pb2


class Offset(
    NamedTuple("NamedTuple", (("element_id", Node), ("displacement", int)))
):
    """
    An Offset describes a location inside a :class:`gtirb.Node`, such as a
    :class:`gtirb.DataBlock` or :class:`gtirb.ByteInterval`.

    :ivar ~.element_id: The :class:`gtirb.Node` containing the location of
            interest.
    :ivar ~.displacement: The offset inside the Node to point to.
    """

    @classmethod
    def _from_protobuf(cls, offset, ir):
        # type: (Offset_pb2.Offset) -> Offset
        """Decode a Protobuf object to an offset.

        :param offset: The Protobuf object.
        """

        element_id = UUID(bytes=offset.element_id)
        return cls(ir.get_by_uuid(element_id), offset.displacement)

    def _to_protobuf(self):
        # type: () -> Offset_pb2.Offset
        """Encode this offset into a Protobuf object."""

        proto_offset = Offset_pb2.Offset()
        proto_offset.element_id = self.element_id.uuid.bytes
        proto_offset.displacement = self.displacement
        return proto_offset
